"""Dataframes from Structure object."""

from PDBNucleicAcids._polymer_dataframe import get_polymer_dataframe
from PDBNucleicAcids._basepair_dataframe import get_basepair_dataframe

__all__ = ["get_basepair_dataframe", "get_polymer_dataframe"]



import pandas as pd
from Bio.PDB.Structure import Structure

# absolute import
from PDBNucleicAcids.utils import get_paired_segments
from PDBNucleicAcids.Pairing import search_paired_base
from PDBNucleicAcids._polymer_dataframe import get_polymer_dataframe


def get_basepair_dataframe(
    structure: Structure,
    nucleic_chain_ids: list[str] | None = None,
) -> pd.DataFrame:
    """
    Return dataframe with base pairs data from a Biopython structure object.

    Parameters
    ----------
    structure : Bio.PDB.Structure.Structure
        Biopython structure object.
    nucleic_chains : list[str] | None, optional
        List that contain chain ids of nucleic acid polymers. If `None` then
        they will be inferred using `polymer_dataframe_from_structure`.

    Returns
    -------
    pandas.DataFrame
        Dataframe with base pairs data in the structure.

    """
    # in case no nucleic chain ids were given
    # get nucleic chains using get_polymer_dataframe
    if nucleic_chain_ids is None:
        # polymer chain data
        polymer_df: pd.DataFrame = get_polymer_dataframe(structure)

        # select only nucleic chains
        nucleic_chain_ids: list[str] = polymer_df[
            polymer_df["polymer_type"] == "polymer_nucleic"
        ]["chain_id"].tolist()

    # list of all nucleic atoms, compute only once
    nucleic_atoms: list = [
        atom
        for atom in structure.get_atoms()
        if atom.parent.parent.id in nucleic_chain_ids
    ]

    # output data
    data: list[tuple[str, int]] = []

    for chain in structure.get_chains():
        if chain.id not in nucleic_chain_ids:
            continue
        for residue in chain:
            i_residue = residue
            j_residue = search_paired_base(
                i_residue, nucleic_atoms=nucleic_atoms
            )

            # in this case there is no base pair
            if j_residue is None:
                continue

            row: tuple[str, int] = (
                i_residue.parent.id,
                i_residue.id[1],
                i_residue.resname,
                j_residue.resname,
                j_residue.id[1],
                j_residue.parent.id,
            )

            # Look if it's on the data list
            if row not in data and row[::-1] not in data:
                data.append(row)

    base_pairs_df = pd.DataFrame(
        data,
        columns=[
            "i_chain_id",
            "i_residue_index",
            "i_residue_name",
            "j_residue_name",
            "j_residue_index",
            "j_chain_id",
        ],
    )

    # Add a column with an index that indicates a paired segment
    # i.e.
    # A  T  0
    # C  G  0
    # C  G  0
    # T  A  1
    base_pairs_df["paired_segment"] = get_paired_segments(base_pairs_df)

    return base_pairs_df



"""Functions that address atom geometry."""

import numpy as np

# from Bio.PDB.Atom import Atom


def calculate_normal_vector(v1, v2, v3) -> np.float64:
    """Compute normal vector from three 3D points."""
    u1 = v2 - v1
    u2 = v3 - v1

    # Calcolo del vettore normale
    normal = np.cross(u1, u2)

    # Normalizzazione manuale del vettore normale
    norm = np.linalg.norm(normal)
    if norm == 0:
        raise ValueError("Vettore normale nullo")
    normal_normalized = normal / norm

    return normal_normalized


def angle_between_vectors(v1, v2) -> np.float64:
    """Compute angle between two vectors."""
    cos_theta = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))

    return np.degrees(np.arccos(np.clip(cos_theta, -1.0, 1.0)))


def angle_between_planes(plane1, plane2) -> np.float64:
    """Compute angle between two planes, composed of three 3D points each."""
    v1, v2, v3 = plane1
    v4, v5, v6 = plane2

    normal1 = calculate_normal_vector(v1, v2, v3)
    normal2 = calculate_normal_vector(v4, v5, v6)

    angle = angle_between_vectors(normal1, normal2)

    return angle


def plane_separation(plane1, plane2) -> np.float64:
    """Calcola la distanza verticale tra due piani."""
    v1, v2, v3 = plane1
    v4, v5, v6 = plane2

    # Ottieni i vettori normali
    normal1 = calculate_normal_vector(v1, v2, v3)
    # normal2 = calculate_normal_vector(v4, v5, v6)

    # Ottieni i centri dei piani
    center1 = np.mean(plane1, axis=0)
    center2 = np.mean(plane2, axis=0)

    # Calcola il vettore tra i centri dei due piani
    center_vector = center2 - center1

    # Proietta il vettore tra i centri lungo la direzione del normale del
    # primo piano
    separation: np.float64 = abs(np.dot(center_vector, normal1))

    return separation


# def angle_between_3_atoms(atom1: Atom, atom2: Atom, atom3: Atom) -> float:
#     """
#     Get angle in degrees between three atoms. OBSOLETE.

#     Atom 1 has the angle that is returned

#     Parameters
#     ----------
#     atom1 : Bio.PDB.Atom.Atom
#         First atom, on which the returned angle rests.
#     atom2 : Bio.PDB.Atom.Atom
#         Second atom.
#     atom3 : Bio.PDB.Atom.Atom
#         Third atom.

#     Returns
#     -------
#     float
#         Angle between the three atoms, measured in degrees.

#     """
#     # get coordinate vectors
#     p1: np.array = atom1.coord
#     p2: np.array = atom2.coord
#     p3: np.array = atom3.coord

#     # Calculate the vectors
#     v1 = p2 - p1  # from p1 to p2
#     v2 = p3 - p1  # from p1 to p3

#     # Calculate the angle in radians
#     # rule of the cosine
#     angle_radians = np.arccos(
#         np.clip(
#             np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)),
#             -1.0,
#             1.0,
#         )
#     )

#     # Convert to degrees, cast into float
#     angle_degrees = np.degrees(angle_radians)
#     angle_degrees = float(angle_degrees)

#     return angle_degrees


# def dihedral_angle_between_4_atoms(
#     atom1: Atom, atom2: Atom, atom3: Atom, atom4: Atom
# ) -> float:
#     """
#     Get dihedral angle in degrees between four atoms. OBSOLETE.

#     The first plane rests of the dihedral angle rests on the first three
#     atoms while the second rests on the last three. `atom1` and `atom2`
#     are in common between the two planes.

#     Parameters
#     ----------
#     atom1 : Bio.PDB.Atom.Atom
#         First atom.
#     atom2 : Bio.PDB.Atom.Atom
#         Second atom.
#     atom3 : Bio.PDB.Atom.Atom
#         Third atom.
#     atom4 : Bio.PDB.Atom.Atom
#         Forth atom.

#     Returns
#     -------
#     float
#         Dihedral angle between the four atoms, measured in degrees.

#     """
#     # Punti tridimensionali, 4 atomi
#     p1 = atom1.coord
#     p2 = atom2.coord
#     p3 = atom3.coord
#     p4 = atom4.coord

#     # Vettori tra i punti
#     v1 = p2 - p1
#     v2 = p3 - p2
#     v3 = p4 - p3

#     # Vettori normali ai piani
#     n1 = np.cross(v1, v2)
#     n2 = np.cross(v2, v3)

#     # Vettore di riferimento per la direzione
#     m = np.cross(n1, v2)

#     # Normalizzazione dei vettori
#     n1 /= np.linalg.norm(n1)
#     n2 /= np.linalg.norm(n2)
#     m /= np.linalg.norm(m)

#     # Calcola coseno e seno dell'angolo
#     cos_theta = np.dot(n1, n2)
#     sin_theta = np.dot(m, n2)

#     # Calcola l'angolo diedro
#     theta = np.arctan2(sin_theta, cos_theta)

#     # Converti l'angolo da radianti a gradi e castalo in float
#     theta_deg = np.degrees(theta)
#     theta_deg = float(theta_deg)

#     return theta_deg



"""Testing the Structure2DataFrame module from PDBNucleicAcids."""

# import pytest
from Bio.PDB.MMCIFParser import MMCIFParser

# to be tested
from PDBNucleicAcids.Structure2DataFrame import get_basepair_dataframe


def get_test_structure():
    filepath = "tests/data/gattaca.cif"
    parser = MMCIFParser(QUIET=True)
    structure = parser.get_structure("gattaca", filepath)

    return structure


def test_invalid_residue():
    """Test per verificare il comportamento con residuo non nucleico."""
    structure = get_test_structure()

    df = get_basepair_dataframe(structure)
    print(df)

    assert True



"""Utils."""

import pandas as pd


def get_paired_segments(df: pd.DataFrame) -> list[int]:
    """
    Get paired segments indexes.

    A - T      0
    T - A      0
    A - T      0
    C   C  ->       ->  [0, 0, 0, 1, 1, 1]
    T   A
    C - G      1
    G - C      1
    T - A      1

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with basepairs data.

    Returns
    -------
    list[int]
        An array of integers of the same length of the input dataframe.
        Integers identify the uninterrupted segments of basepairs.

    """
    # trivially true
    # if there is only one pair base that there is only one paired segment
    if len(df) == 1:
        return [0]

    paired_segment_ids: list[int] = []

    # initialized data from previous bp and current bp as None
    previous_i_chain_id: str | None = None
    previous_i_residue_index: int | None = None
    previous_j_chain_id: str | None = None
    previous_j_residue_index: int | None = None

    current_paired_segment = 0  # the first paired segment is index 0

    current_i_chain_id: str | None = None
    current_i_residue_index: int | None = None
    current_j_chain_id: str | None = None
    current_j_residue_index: int | None = None

    for i, row in df.iterrows():
        # update data for the current base pair (row : base pair)
        current_i_chain_id: str = row["i_chain_id"]
        current_i_residue_index: int = row["i_residue_index"]
        current_j_chain_id: str = row["j_chain_id"]
        current_j_residue_index: int = row["j_residue_index"]

        if previous_i_chain_id is None:
            # starting case, the very first base pair encountered
            paired_segment_ids.append(current_paired_segment)

        elif (
            previous_i_chain_id == current_i_chain_id
            and previous_j_chain_id == current_j_chain_id
            and previous_i_residue_index + 1 == current_i_residue_index
            and previous_j_residue_index - 1 == current_j_residue_index
        ):
            # step case, all the other base pairs encountered
            # continuity case
            paired_segment_ids.append(current_paired_segment)
        else:
            # step case, all the other base pairs encountered
            # discontinuity case
            current_paired_segment += 1
            paired_segment_ids.append(current_paired_segment)

        # update data for the previous base pair (row : base pair)
        previous_i_chain_id = current_i_chain_id
        previous_i_residue_index = current_i_residue_index
        previous_j_chain_id = current_j_chain_id
        previous_j_residue_index = current_j_residue_index

    # an array of segment indexes
    return paired_segment_ids



"""Classes with rules for proper base pairing."""

import warnings

# from Bio.PDB.vectors import calc_dihedral
from Bio.PDB.Residue import Residue
from Bio.PDB.Atom import Atom
from Bio.PDB.Polypeptide import is_nucleic

# absolute imports
from PDBNucleicAcids.Geometry import angle_between_planes, plane_separation


class dsDNAWatsonCrickBasePairRules:
    """
    Rules for Watson-Crick base pairs in double-strand DNA.

    Parameters
    ----------
    max_distance : float | int, optional
        Maximum distance between nucleotide bases, in Armstrong.
        The default is 3.5.
    max_angle : float | int, optional
        Maximum angle between the planes of the bases, in degrees.
        The default is 65.
    max_stagger : float | int, optional
        Maximum vertical distance between the planes of the bases, in degrees.
        The default is 65.

    """

    def __init__(
        self,
        max_distance: float | int = 4,
        max_angle: float | int = 65,
        max_stagger: float | int = 2.5,
        # dihedral_range: tuple[int, float] = (90, 150),
    ):
        self.complementary_pairs: set[tuple[str]] = {
            ("DA", "DT"),
            ("DT", "DA"),
            ("DG", "DC"),
            ("DC", "DG"),
        }

        # parameters taken from
        # http://forum.x3dna.org/faqs/
        # how-to-fix-missing-(superfluous)-base-pairs-identified-by-find_pair/
        self.max_distance = max_distance
        self.max_angle = max_angle
        self.max_stagger = max_stagger
        # self.dihedral_angle_range = dihedral_range

    def atoms_from_base_ring(self, base: Residue) -> tuple[Atom] | None:
        """Get atoms from the ring of a base."""
        if base.get_resname() in ["DA", "DG"]:  # Purine
            atoms: tuple[Atom] = (base["N1"], base["C2"], base["C6"])
        elif base.get_resname() in ["DT", "DC"]:  # Pirimidine
            atoms: tuple[Atom] = (base["N3"], base["C2"], base["C4"])
        elif base.get_resname() == "HOH":  # Water
            atoms = None
        else:  # Hetero-residue?
            warnings.warn(
                f"{base.full_id} has no atoms from base ring. Might \
be hetero-residue non-standard-base."
            )
            atoms = None
        return atoms

    def atom_from_central_hbond(self, base) -> Atom | None:
        """Get central H-bond atom of a base."""
        atoms = self.atoms_from_base_ring(base)
        if atoms:
            return atoms[0]
        else:
            return None

    def distance(self, base1, base2) -> float | None:
        """Compute distance between central H-bond atoms of two bases."""
        atom1 = self.atom_from_central_hbond(base1)
        atom2 = self.atom_from_central_hbond(base2)
        if atom1 is not None and atom2 is not None:
            return atom1 - atom2
        else:
            return None

    def angle_between_bases(
        self, base1: Residue, base2: Residue
    ) -> float | None:
        """Compute angle between between planes of two bases."""
        atom_list1 = self.atoms_from_base_ring(base1)
        atom_list2 = self.atoms_from_base_ring(base2)
        if atom_list1 is None or atom_list2 is None:
            return None
        plane1 = (atom.coord for atom in atom_list1)
        plane2 = (atom.coord for atom in atom_list2)
        return angle_between_planes(plane1, plane2)

    def stagger(self, base1: Residue, base2: Residue) -> float | None:
        """Compute vertical stagger between between planes of two bases."""
        atom_list1 = self.atoms_from_base_ring(base1)
        atom_list2 = self.atoms_from_base_ring(base2)
        if atom_list1 is None or atom_list2 is None:
            return None
        plane1 = (atom.coord for atom in atom_list1)
        plane2 = (atom.coord for atom in atom_list2)
        return plane_separation(plane1, plane2)

    # def dihedrals(self, base1, base2) -> float:
    #     atom1, atom2, atom3 = self.atoms_from_base_ring(base1)
    #     atom4, atom5, atom6 = self.atoms_from_base_ring(base2)

    #     v1, v2, v3, v4, v5, v6 = (
    #         atom1.get_vector(),
    #         atom2.get_vector(),
    #         atom3.get_vector(),
    #         atom4.get_vector(),
    #         atom5.get_vector(),
    #         atom6.get_vector(),
    #     )
    #     dihedral1 = calc_dihedral(v1, v2, v3, v4)
    #     dihedral2 = calc_dihedral(v1, v5, v6, v4)

    #     return (dihedral1, dihedral2)

    def is_different_residue(self, base1: Residue, base2: Residue) -> bool:
        """Check if two bases are distinct bases."""
        return base1 != base2

    def is_complementary(self, base1: Residue, base2: Residue) -> bool:
        """Check if two bases are complementary."""
        pair: tuple[str] = (base1.get_resname(), base2.get_resname())
        return pair in self.complementary_pairs

    def is_from_different_chain(self, base1: Residue, base2: Residue) -> bool:
        """Check if two bases are from distinct chains."""
        return base1.parent != base2.parent

    def is_valid_distance(self, base1: Residue, base2: Residue) -> bool:
        """Check validity of distance between central two bases."""
        dist = self.distance(base1, base2)
        if dist is not None:
            return dist <= self.max_distance
        else:
            return False

    def is_valid_angle(self, base1: Residue, base2: Residue) -> bool:
        """Check validity of angle between between planes of the bases."""
        angle = self.angle_between_bases(base1, base2)
        if angle is not None:
            return angle <= self.max_angle or angle >= 180 - self.max_angle
        else:
            False

    def is_valid_stagger(self, base1: Residue, base2: Residue) -> bool:
        """Check validity of stagger between between planes of the bases."""
        # TODO to be used
        return self.stagger(base1, base2) <= self.max_stagger

    def is_candidate(self, base1: Residue, base2: Residue) -> bool:
        """
        Check if (base1, base2) is a likely base pair.

        Use all the constraint methods to infer if base2 is likely to be
        paired with base1.
        """
        if not is_nucleic(base1, standard=False):
            # TODO add warning
            return False
        elif not is_nucleic(base2, standard=False):
            return False
        elif not is_nucleic(base1, standard=True):
            # TODO add warning
            return False
        elif not is_nucleic(base2, standard=True):
            return False
        elif not self.is_different_residue(base1, base2):
            return False
        elif not self.is_complementary(base1, base2):
            return False
        elif not self.is_from_different_chain(base1, base2):
            return False
        elif not self.is_valid_distance(base1, base2):
            return False
        elif not self.is_valid_angle(base1, base2):
            return False
        else:
            return True



"""Single base pairing."""

from Bio.PDB.Structure import Structure
from Bio.PDB.Residue import Residue
from Bio.PDB.Atom import Atom
from Bio.PDB.NeighborSearch import NeighborSearch

# absolute imports
from PDBNucleicAcids._polymer_dataframe import get_polymer_dataframe
from PDBNucleicAcids.BasePairRules import dsDNAWatsonCrickBasePairRules


def search_paired_base(
    residue: Residue,
    nucleic_chain_ids: list[str] | None = None,
    nucleic_atoms: list[Atom] | None = None,
    pairing_rules=dsDNAWatsonCrickBasePairRules,
) -> Residue | None:
    """
    Search in the vicinity of a given base for its paired base.

    Parameters
    ----------
    residue : Bio.PDB.Residue.Residue
        A Biopython nucleic acid residue (nucleotide) taken from a Biopython
        structure.
    nucleic_chain_ids : list[str], optional
        List of ids for nucleic acid chains. If None they will be
        inferred using polymer_dataframe_from_structure(). Default is None.
    nucleic_atoms : list[Atom], optional
        List of atoms from the nucleic acid chains. If None they will be
        inferred using nucleic_chain_ids(). Default is None.
    pairing_rules : optional
        Class instance with rules for proper pairing.
        PDBNucleicAcids.BasePairsRules.WatsonCrickBasePairsRules
        is the default. Currently it might work only with those rules

    Returns
    -------
    Bio.PDB.Residue.Residue | None
        Nucleotide that binds to the input nucleotide residue.
        None in the case there is no nucleotide paired to the input nucleotide.

    """
    # from residue get to the structure
    structure: Structure = residue.parent.parent.parent

    # in case no nucleic atoms and no nucleic chain ids were given
    # get nucleic chains using get_polymer_dataframe
    if not nucleic_atoms and not nucleic_chain_ids:
        polymer_df = get_polymer_dataframe(structure)
        nucleic_chain_ids = polymer_df[
            polymer_df["polymer_type"] == "polymer_nucleic"
        ]["chain_id"].tolist()

    # in case no nucleic atoms were given but
    # nucleic chain ids were given or got from get_polymer_dataframe
    # list of all nucleic atoms, compute only once
    if not nucleic_atoms and nucleic_chain_ids:
        nucleic_atoms: list[Atom] = [
            atom
            for atom in structure.get_atoms()
            if atom.parent.parent.id in nucleic_chain_ids
            and atom.parent != residue  # not from the same residue
        ]

    # if there are no nucleic atoms except from the one input residue
    # then NeighborSearch will return an error
    if len(nucleic_atoms) == 0 or nucleic_atoms is None:
        return None

    # initialize NeighborSearch class
    ns = NeighborSearch(nucleic_atoms)

    # initialize Rules class
    pairing_rules = pairing_rules()

    central_atom: Atom = pairing_rules.atom_from_central_hbond(residue)
    # if the input residue is not a DNA base then the central atom
    # will be None
    if central_atom is None:
        return None

    # search around the central atom of the residue
    atom_neighborhood = ns.search(center=central_atom.coord, radius=4.0)

    residue_neighborhood: list[Residue] = list(
        {atom.parent for atom in atom_neighborhood}
    )

    candidate_residues: list[Residue] = []

    for residue_neighbour in residue_neighborhood:
        if pairing_rules.is_candidate(base1=residue, base2=residue_neighbour):
            dist = pairing_rules.distance(
                base1=residue, base2=residue_neighbour
            )
            candidate_residues.append((residue_neighbour, dist))

    if len(candidate_residues) > 1:
        # If there is more than one paired residue
        min_tuple: tuple = min(candidate_residues, key=lambda x: x[1])
        return min_tuple[0]
    elif candidate_residues:
        # only one paired nucleotide found
        return candidate_residues[0][0]
    else:
        # no paired nucleotides found
        return None



# import warnings
import pandas as pd
from Bio.PDB.Polypeptide import is_nucleic
from Bio.PDB.Polypeptide import is_aa
from Bio.PDB.Structure import Structure


def get_polymer_dataframe(structure: Structure) -> pd.DataFrame:
    """
    Return dataframe with polymer data from a Biopython structure object.

    Parameters
    ----------
    structure : Bio.PDB.Structure.Structure
        Biopython structure object.

    Returns
    -------
    pandas.DataFrame
        Dataframe with polymer data in the structure.

    """
    data = []

    for chain in structure.get_chains():
        nucleotide_counter = 0
        aa_counter = 0
        water_counter = 0
        other_counter = 0

        for residue in chain:
            # it's a nucleotide
            # TODO distinguish between DNA and RNA
            if is_nucleic(residue):
                nucleotide_counter += 1
            # it's an aminoacid
            elif is_aa(residue):
                aa_counter += 1
            # it's water
            elif residue.resname == "HOH":
                water_counter += 1
            # it's something else
            else:
                other_counter += 1

        # stats = (nucleotide_counter, aa_counter,
        #          water_counter, other_counter)
        tot = nucleotide_counter + aa_counter + other_counter

        if nucleotide_counter / tot > 0.7:
            data.append((chain.id, "polymer_nucleic"))
        elif aa_counter / tot > 0.7:
            data.append((chain.id, "polymer_protein"))
        else:
            # warnings.warn(
            # f"Unknown polymer. id = {structure.id}, chain = \
            # {chain.id}, stats = {stats}"
            #             )
            data.append((chain.id, "unknown"))

    df = pd.DataFrame(data, columns=["chain_id", "polymer_type"])

    return df
    # TODO
    # "non_standard_linkage": non_standard_linkage,
    # "non_standard_residue": non_standard_residue,



"""Extract basepair data from the NDB/NAKB Data Category in mmCif files."""

from pathlib import Path
import pandas as pd
from Bio.PDB.MMCIF2Dict import MMCIF2Dict

# absolute import
from PDBNucleicAcids.utils import get_paired_segments


def basepair_dataframe_from_mmcif(
    mmcif_filepath: str | Path,
) -> pd.DataFrame | None:
    """
    Return dataframe with base pairs data from a mmCif file.

    Note: not all mmCif files contain Data Category "ndb_struct_na_base_pair"

    Parameters
    ----------
    mmcif_filepath : str | Path
        mmCif filepath.

    Returns
    -------
    pandas.DataFrame | None
        Dataframe with base pairs data in the structure.

    """
    mmcif_dict = MMCIF2Dict(str(mmcif_filepath))

    try:
        pair_name = mmcif_dict["_ndb_struct_na_base_pair.pair_name"]
        i_chain_id = mmcif_dict["_ndb_struct_na_base_pair.i_auth_asym_id"]
        i_residue_index = mmcif_dict["_ndb_struct_na_base_pair.i_auth_seq_id"]
        i_residue_name = mmcif_dict["_ndb_struct_na_base_pair.i_label_comp_id"]
        j_chain_id = mmcif_dict["_ndb_struct_na_base_pair.j_auth_asym_id"]
        j_residue_index = mmcif_dict["_ndb_struct_na_base_pair.j_auth_seq_id"]
        j_residue_name = mmcif_dict["_ndb_struct_na_base_pair.j_label_comp_id"]
        base_pairs_df = pd.DataFrame(
            {
                "pair_name": pair_name,
                "i_chain_id": i_chain_id,
                "i_residue_index": list(map(int, i_residue_index)),
                "i_residue_name": i_residue_name,
                "j_chain_id": j_chain_id,
                "j_residue_index": list(map(int, j_residue_index)),
                "j_residue_name": j_residue_name,
            }
        )
    except KeyError as e:
        # in this case at least one of these keys are not present
        # so there is no basepair datastructure
        print(e)
        return None

    polymers_df = polymer_dataframe_from_mmcif(mmcif_filepath)

    # Divide into two dataframes
    # one has the data from the "i" side and the other from the "j" side
    # of the chain
    i_polymers_df = polymers_df.copy()
    j_polymers_df = polymers_df.copy()

    # renaming columns, adding a prefix "i_" and prefix "j_"
    i_polymers_df.columns = map(lambda col: "i_" + col, i_polymers_df.columns)
    j_polymers_df.columns = map(lambda col: "j_" + col, j_polymers_df.columns)

    # dataframe with base pair information AND polymer information
    # by inneri joining the two dataframes, from "i" and "j"
    result_df = pd.merge(base_pairs_df, i_polymers_df, on="i_chain_id")
    result_df = pd.merge(result_df, j_polymers_df, on="j_chain_id")

    # rearranging columns by slicing the dataframe and then concatenating
    i_result_df = result_df[
        [
            "pair_name",
            "i_polymer_type",
            "i_non_standard_linkage",
            "i_non_standard_residue",
            "i_chain_id",
            "i_residue_index",
            "i_residue_name",
        ]
    ]
    j_result_df = result_df[
        [
            "j_residue_name",
            "j_residue_index",
            "j_chain_id",
            "j_non_standard_residue",
            "j_non_standard_linkage",
            "j_polymer_type",
        ]
    ]
    result_df = pd.concat([i_result_df, j_result_df], axis=1)

    # Add a column with an index that indicates a paired segment
    # i.e.
    # A  T  0
    # C  G  0
    # C  G  0
    # T  A  1
    result_df["paired_segment"] = get_paired_segments(result_df)

    return result_df


def polymer_dataframe_from_mmcif(
    mmcif_filepath: str | Path,
) -> pd.DataFrame | None:
    """
    Dataframe with polymer data from a mmCif file.

    Parameters
    ----------
    mmcif_filepath : str | Path
        mmCif filepath.

    Returns
    -------
    pandas.DataFrame | None
        Dataframe with information about each polymer in the structure.

    """
    # TODO test with weird chain ids like multimeric
    # i.e. "A,B,C,D"
    mmcif_dict = MMCIF2Dict(str(mmcif_filepath))

    try:
        polymer_type = mmcif_dict["_entity_poly.type"]
        non_standard_linkage = mmcif_dict["_entity_poly.nstd_linkage"]
        non_standard_residue = mmcif_dict["_entity_poly.nstd_monomer"]
        polymers_df = pd.DataFrame(
            {
                "polymer_type": polymer_type,
                "non_standard_linkage": non_standard_linkage,
                "non_standard_residue": non_standard_residue,
            }
        )
    except KeyError as e:
        print(e)
        return None

    # there are two options for chain IDs:
    # ndb (nucleotide db) and pdbx (protein db X)
    # sometimes it's called chain, sometimes it's called strand
    if "_entity_poly.pdbx_chain_id" in mmcif_dict:
        polymers_df["chain_id"] = mmcif_dict["_entity_poly.pdbx_chain_id"]
    elif "_entity_poly.pdbx_strand_id" in mmcif_dict:
        polymers_df["chain_id"] = mmcif_dict["_entity_poly.pdbx_strand_id"]
    elif "_entity_poly.ndb_chain_id" in mmcif_dict:
        polymers_df["chain_id"] = mmcif_dict["_entity_poly.ndb_chain_id"]
    elif "_entity_poly.ndb_strand_id" in mmcif_dict:
        polymers_df["chain_id"] = mmcif_dict["_entity_poly.ndb_strand_id"]
    else:
        return None

    return polymers_df



"""Tests for the BasePairRules module from PDBNucleicAcids."""

import pytest
import numpy as np
from unittest.mock import MagicMock
from Bio.PDB.Residue import Residue
from Bio.PDB.Atom import Atom

# to be tested
from PDBNucleicAcids.BasePairRules import dsDNAWatsonCrickBasePairRules


@pytest.fixture
def mock_residue():
    """Fixture for creating a mock residue with specific atoms."""
    residue = MagicMock(spec=Residue)
    residue.get_resname.return_value = "DA"
    residue.atoms = {
        "N1": Atom("N1", np.array([0.0, 0.0, 0.0]), 1.0, 1.0, " ", "N1", 1),
        "C2": Atom("C2", np.array([1.0, 0.0, 0.0]), 1.0, 1.0, " ", "C2", 2),
        "C6": Atom("C6", np.array([0.0, 1.0, 0.0]), 1.0, 1.0, " ", "C6", 3),
    }
    residue.__getitem__.side_effect = residue.atoms.__getitem__

    # Aggiunta del parent come catena mock
    chain = MagicMock()
    chain.id = "A"  # Identificatore della catena
    residue.parent = chain

    return residue


@pytest.fixture
def complementary_residue():
    """Fixture for creating a complementary mock residue."""
    residue = MagicMock(spec=Residue)
    residue.get_resname.return_value = "DT"
    residue.atoms = {
        "N3": Atom("N3", np.array([0.0, 0.0, 3.5]), 1.0, 1.0, " ", "N3", 4),
        "C2": Atom("C2", np.array([1.0, 0.0, 3.5]), 1.0, 1.0, " ", "C2", 5),
        "C4": Atom("C4", np.array([0.0, 1.0, 3.5]), 1.0, 1.0, " ", "C4", 6),
    }
    residue.__getitem__.side_effect = residue.atoms.__getitem__

    # Aggiunta del parent come un'altra catena mock
    chain = MagicMock()
    chain.id = "B"  # Identificatore della catena, diverso da mock_residue
    residue.parent = chain

    return residue


@pytest.fixture
def base_pair_rules():
    """Fixture for the dsDNAWatsonCrickBasePairRules instance."""
    return dsDNAWatsonCrickBasePairRules()


def test_initialization(base_pair_rules):
    """Test to verify base parameters initialization."""
    assert base_pair_rules.max_distance == 4
    assert base_pair_rules.max_angle == 65
    assert base_pair_rules.max_stagger == 2.5
    assert ("DA", "DT") in base_pair_rules.complementary_pairs


def test_is_different_residue(
    base_pair_rules, mock_residue, complementary_residue
):
    """Test to verify same/different bases."""
    # Positive test, two different complementary base
    assert base_pair_rules.is_different_residue(
        mock_residue, complementary_residue
    )

    # same base
    assert not base_pair_rules.is_different_residue(mock_residue, mock_residue)
    assert not base_pair_rules.is_candidate(mock_residue, mock_residue)


def test_is_complementary(
    base_pair_rules, mock_residue, complementary_residue
):
    """Test to verify base complementarity."""
    # Positive test, complementary base
    assert base_pair_rules.is_complementary(
        mock_residue, complementary_residue
    )

    # not complementary base
    complementary_residue.get_resname.return_value = "DG"
    assert not base_pair_rules.is_complementary(
        mock_residue, complementary_residue
    )
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )


def test_is_from_different_chain(
    base_pair_rules, mock_residue, complementary_residue
):
    """Test to verify base complementarity."""
    # Positive test, different chains
    assert base_pair_rules.is_from_different_chain(
        mock_residue, complementary_residue
    )

    # same chain
    complementary_residue.parent = mock_residue.parent

    assert not base_pair_rules.is_from_different_chain(
        mock_residue, complementary_residue
    )
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )


def test_is_valid_distance(
    base_pair_rules, mock_residue, complementary_residue
):
    """Test to verify valid distance between bases."""
    # Positive test
    assert base_pair_rules.is_valid_distance(
        mock_residue, complementary_residue
    )

    # Test with larger distance
    complementary_residue["N3"].coord = np.array([0.0, 0.0, 8.0])

    assert not base_pair_rules.is_valid_distance(
        mock_residue, complementary_residue
    )
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )

    # internal methods return None
    mock_residue.get_resname.return_value = "HOH"
    assert not base_pair_rules.is_valid_distance(
        mock_residue, complementary_residue
    )


def test_is_valid_angle(base_pair_rules, mock_residue, complementary_residue):
    """Test to verify valid angle between bases."""
    # Positive test
    assert base_pair_rules.is_valid_angle(mock_residue, complementary_residue)

    # Out-of-limit angle
    complementary_residue["N3"].coord = np.array([0.0, 0.0, 1.0])

    assert not base_pair_rules.is_valid_angle(
        mock_residue, complementary_residue
    )
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )

    # internal methods return None
    mock_residue.get_resname.return_value = "HOH"
    assert not base_pair_rules.is_valid_angle(
        mock_residue, complementary_residue
    )


def test_is_candidate(base_pair_rules, mock_residue, complementary_residue):
    """Test to verify if a pair of bases meets candidate criteria."""
    # Positive test
    assert base_pair_rules.is_candidate(mock_residue, complementary_residue)

    # non-nucleic residue test
    mock_residue.get_resname.return_value = "LYS"
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )

    # non-nucleic residue test
    mock_residue.get_resname.return_value = "DA"
    complementary_residue.get_resname.return_value = "LYS"
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )

    # non-standard nucleic base test
    mock_residue.get_resname.return_value = "A2L"
    complementary_residue.get_resname.return_value = "DA"
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )

    # non-standard nucleic base test
    mock_residue.get_resname.return_value = "DA"
    complementary_residue.get_resname.return_value = "A2L"
    assert not base_pair_rules.is_candidate(
        mock_residue, complementary_residue
    )
